/**
 * @file Services are a special kind of caveat based off of
 * the official lsat spec by
 * [Lightning Labs](https://lsat.tech/macaroons#target-services).
 * These have certain expectations around value encoding and also support
 * tiers of capabilities where services have service level capabilities and
 * capabilities in turn can have constraints.
 * See below for an example from lightning loop.
 *
 * @example
 *  services = lightning_loop:0
 *  lightning_loop_capabilities = loop_out,loop_in
 *  loop_out_monthly_volume_sats = 200000000
 *
 */
import bufio from 'bufio';
import { Caveat } from './caveat';
export declare class NoServicesError extends Error {
    constructor(...params: string[]);
}
export declare class InvalidServicesError extends Error {
    constructor(message?: string);
}
export declare class InvalidCapabilitiesError extends Error {
    constructor(message?: string);
}
export interface ServiceClassOptions {
    name: string;
    tier: number;
}
export declare class Service extends bufio.Struct {
    name: string;
    tier: number;
    constructor(options: ServiceClassOptions);
}
export declare const SERVICES_CAVEAT_CONDITION = "services";
/**
 *
 * @param {string} s - raw services string of format `name:tier,name:tier`
 * @returns array of Service objects or throws an error
 */
export declare const decodeServicesCaveat: (s: string) => Service[] | Error;
export declare const encodeServicesCaveatValue: (services: Service[]) => string | Error;
export declare const SERVICE_CAPABILITIES_SUFFIX = "_capabilities";
export declare const createNewCapabilitiesCaveat: (serviceName: string, _capabilities?: string | string[] | undefined) => Caveat;
export declare const decodeCapabilitiesValue: (value: string) => string[];
